
#include <errno.h>
#include <stdlib.h>

#include <sstream>

#include "common/armor.h"
#include "common/Clock.h"

#include "rgw_access.h"
#include "rgw_op.h"
#include "rgw_rest.h"
#include "rgw_acl.h"
#include "rgw_user.h"
#include "rgw_log.h"
#include "rgw_multi.h"

using namespace std;
using ceph::crypto::MD5;

static string mp_ns = "multipart";
static string tmp_ns = "tmp";

class MultipartMetaFilter : public RGWAccessListFilter {
public:
  MultipartMetaFilter() {}
  bool filter(string& name, string& key) {
    int len = name.size();
    if (len < 6)
      return false;

    int pos = name.find(MP_META_SUFFIX, len - 5);
    if (pos <= 0)
      return false;

    pos = name.rfind('.', pos - 1);
    if (pos < 0)
      return false;

    key = name.substr(0, pos);

    return true;
  }
};

static MultipartMetaFilter mp_filter;

static int parse_range(const char *range, off_t& ofs, off_t& end)
{
  int r = -ERANGE;
  string s(range);
  int pos = s.find("bytes=");
  string ofs_str;
  string end_str;

  if (pos < 0)
    goto done;

  s = s.substr(pos + 6); /* size of("bytes=")  */
  pos = s.find('-');
  if (pos < 0)
    goto done;

  ofs_str = s.substr(0, pos);
  end_str = s.substr(pos + 1);
  if (ofs_str.length())
    ofs = atoll(ofs_str.c_str());

  if (end_str.length())
  end = atoll(end_str.c_str());

  RGW_LOG(10) << "parse_range ofs=" << ofs << " end=" << end << dendl;

  if (end < ofs)
    goto done;

  r = 0;
done:
  return r;
}

/**
 * Get the HTTP request metadata out of the req_state as a
 * map(<attr_name, attr_contents>, where attr_name is RGW_ATTR_PREFIX.HTTP_NAME)
 * s: The request state
 * attrs: will be filled up with attrs mapped as <attr_name, attr_contents>
 *
 */
void get_request_metadata(struct req_state *s, map<string, bufferlist>& attrs)
{
  map<string, string>::iterator iter;
  for (iter = s->x_amz_map.begin(); iter != s->x_amz_map.end(); ++iter) {
    string name = iter->first;
#define X_AMZ_META "x-amz-meta"
    if (name.find(X_AMZ_META) == 0) {
      RGW_LOG(10) << "x>> " << iter->first << ":" << iter->second << dendl;
      string& val = iter->second;
      bufferlist bl;
      bl.append(val.c_str(), val.size() + 1);
      string attr_name = RGW_ATTR_PREFIX;
      attr_name.append(name);
      attrs[attr_name.c_str()] = bl;
    }
  }
}

/**
 * Get the AccessControlPolicy for an object off of disk.
 * policy: must point to a valid RGWACL, and will be filled upon return.
 * bucket: name of the bucket containing the object.
 * object: name of the object to get the ACL for.
 * Returns: 0 on success, -ERR# otherwise.
 */
static int get_policy_from_attr(RGWAccessControlPolicy *policy, rgw_obj& obj)
{
  bufferlist bl;
  int ret = 0;

  if (obj.bucket.size()) {
    ret = rgwstore->get_attr(obj, RGW_ATTR_ACL, bl);

    if (ret >= 0) {
      bufferlist::iterator iter = bl.begin();
      policy->decode(iter);
      if (g_conf->rgw_log >= 15) {
        RGW_LOG(15) << "Read AccessControlPolicy" << dendl;
        policy->to_xml(cerr);
        RGW_LOG(15) << dendl;
      }
    }
  }

  return ret;
}

int read_acls(struct req_state *s, RGWAccessControlPolicy *policy, string& bucket, string& object)
{
  string upload_id = s->args.get("uploadId");
  string oid = object;
  rgw_obj obj;

  if (!oid.empty()) {
    bool suspended;
    int ret = rgwstore->bucket_suspended(bucket, &suspended);
    if (ret < 0)
      return ret;

    if (suspended)
      return -ERR_USER_SUSPENDED;
  }

  if (!oid.empty() && !upload_id.empty()) {
    RGWMPObj mp(oid, upload_id);
    oid = mp.get_meta();
    obj.set_ns(mp_ns);
  }
  obj.init(bucket, oid, object);

  int ret = get_policy_from_attr(policy, obj);
  if (ret == -ENOENT && object.size()) {
    /* object does not exist checking the bucket's ACL to make sure
       that we send a proper error code */
    RGWAccessControlPolicy bucket_policy;
    string no_object;
    rgw_obj no_obj(bucket, no_object);
    ret = get_policy_from_attr(&bucket_policy, no_obj);
    if (ret < 0)
      return ret;

    if (!verify_permission(&bucket_policy, s->user.user_id, s->perm_mask, RGW_PERM_READ))
      ret = -EACCES;
    else
      ret = -ENOENT;
  } else if (ret == -ENOENT) {
      ret = -ERR_NO_SUCH_BUCKET;
  }

  return ret;
}

/**
 * Get the AccessControlPolicy for a bucket or object off of disk.
 * s: The req_state to draw information from.
 * only_bucket: If true, reads the bucket ACL rather than the object ACL.
 * Returns: 0 on success, -ERR# otherwise.
 */
int read_acls(struct req_state *s, bool only_bucket)
{
  int ret = 0;
  string obj_str;

  if (!s->acl) {
     s->acl = new RGWAccessControlPolicy;
     if (!s->acl)
       return -ENOMEM;
  }

  /* we're passed only_bucket = true when we specifically need the bucket's
     acls, that happens on write operations */
  if (!only_bucket)
    obj_str = s->object_str;

  ret = read_acls(s, s->acl, s->bucket_str, obj_str);

  return ret;
}

int RGWGetObj::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_READ))
    return -EACCES;

  return 0;
}

void RGWGetObj::execute()
{
  void *handle = NULL;
  rgw_obj obj;

  ret = get_params();
  if (ret < 0)
    goto done;

  init_common();

  obj.init(s->bucket_str, s->object_str);
  ret = rgwstore->prepare_get_obj(obj, ofs, &end, &attrs, mod_ptr,
                                  unmod_ptr, &lastmod, if_match, if_nomatch, &total_len, &handle, &s->err);
  if (ret < 0)
    goto done;

  s->obj_size = total_len;

  if (!get_data || ofs > end)
    goto done;

  while (ofs <= end) {
    ret = rgwstore->get_obj(&handle, obj, &data, ofs, end);
    if (ret < 0) {
      goto done;
    }
    len = ret;
    ofs += len;
    ret = 0;

    send_response(handle);
    free(data);
  }

  return;

done:
  send_response(handle);
  free(data);
  rgwstore->finish_get_obj(&handle);
}

int RGWGetObj::init_common()
{
  if (range_str) {
    int r = parse_range(range_str, ofs, end);
    if (r < 0)
      return r;
  }
  if (if_mod) {
    if (parse_time(if_mod, &mod_time) < 0)
      return -EINVAL;
    mod_ptr = &mod_time;
  }

  if (if_unmod) {
    if (parse_time(if_unmod, &unmod_time) < 0)
      return -EINVAL;
    unmod_ptr = &unmod_time;
  }

  return 0;
}

int RGWListBuckets::verify_permission()
{
  return 0;
}

void RGWListBuckets::execute()
{
  ret = rgw_read_user_buckets(s->user.user_id, buckets, !!(s->prot_flags & RGW_REST_OPENSTACK));
  if (ret < 0) {
    /* hmm.. something wrong here.. the user was authenticated, so it
       should exist, just try to recreate */
    RGW_LOG(10) << "WARNING: failed on rgw_get_user_buckets uid=" << s->user.user_id << dendl;

    /*

    on a second thought, this is probably a bug and we should fail

    rgw_put_user_buckets(s->user.user_id, buckets);
    ret = 0;

    */
  }

  send_response();
}

int RGWStatBucket::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_READ))
    return -EACCES;

  return 0;
}

void RGWStatBucket::execute()
{
  RGWUserBuckets buckets;
  bucket.name = s->bucket;
  buckets.add(bucket);
  map<string, RGWBucketEnt>& m = buckets.get_buckets();
  ret = rgwstore->update_containers_stats(m);
  if (!ret)
    ret = -EEXIST;
  if (ret > 0) {
    ret = 0;
    map<string, RGWBucketEnt>::iterator iter = m.find(bucket.name);
    if (iter != m.end()) {
      bucket = iter->second;
    } else {
      ret = -EINVAL;
    }
  }

  send_response();
}

int RGWListBucket::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_READ))
    return -EACCES;

  return 0;
}

void RGWListBucket::execute()
{
  string no_ns;

  url_decode(s->args.get("prefix"), prefix);
  marker = s->args.get("marker");
  max_keys = s->args.get(limit_opt_name);
  if (!max_keys.empty()) {
    max = atoi(max_keys.c_str());
  } else {
    max = default_max;
  }
  url_decode(s->args.get("delimiter"), delimiter);

  if (s->prot_flags & RGW_REST_OPENSTACK) {
    string path_args = s->args.get("path");
    if (!path_args.empty()) {
      if (!delimiter.empty() || !prefix.empty()) {
        ret = -EINVAL;
        goto done;
      }
      url_decode(path_args, prefix);
      delimiter="/";
    }
  }

  ret = rgwstore->list_objects(s->user.user_id, s->bucket_str, max, prefix, delimiter, marker, objs, common_prefixes,
                               !!(s->prot_flags & RGW_REST_OPENSTACK), no_ns, NULL, NULL);

done:
  send_response();
}

int RGWCreateBucket::verify_permission()
{
  if (!rgw_user_is_authenticated(s->user))
    return -EACCES;

  return 0;
}

void RGWCreateBucket::execute()
{
  RGWAccessControlPolicy policy, old_policy;
  map<string, bufferlist> attrs;
  bufferlist aclbl;
  bool existed;
  bool pol_ret;
  int pool_id;

  rgw_obj obj(rgw_root_bucket, s->bucket_str);

  int r = get_policy_from_attr(&old_policy, obj);
  if (r >= 0)  {
    if (old_policy.get_owner().get_id().compare(s->user.user_id) != 0) {
      ret = -EEXIST;
      goto done;
    }
  }

  pol_ret = policy.create_canned(s->user.user_id, s->user.display_name, s->canned_acl);
  if (!pol_ret) {
    ret = -EINVAL;
    goto done;
  }
  policy.encode(aclbl);

  attrs[RGW_ATTR_ACL] = aclbl;

  ret = rgwstore->create_bucket(s->user.user_id, s->bucket_str, attrs, true,
				s->user.auid);
  /* continue if EEXIST and create_bucket will fail below.  this way we can recover
   * from a partial create by retrying it. */
  if (ret && ret != -EEXIST)   
    goto done;

  existed = (ret == -EEXIST);

  ret = rgw_add_bucket(s->user.user_id, s->bucket_str);
  if (ret && !existed && ret != -EEXIST)   /* if it exists (or previously existed), don't remove it! */
    rgw_remove_bucket(s->user.user_id, s->bucket_str, false);

  if (ret == -EEXIST)
    ret = 0;

  pool_id = rgwstore->get_bucket_id(s->bucket_str);
  if (pool_id >= 0) {
    s->pool_id = pool_id;
    RGWPoolInfo info;
    info.owner = s->user.user_id;
    info.bucket = s->bucket_str;
    rgw_store_pool_info(pool_id, info);
  }

done:
  send_response();
}

int RGWDeleteBucket::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_WRITE))
    return -EACCES;

  return 0;
}

void RGWDeleteBucket::execute()
{
  ret = -EINVAL;

  if (s->bucket) {
    ret = rgwstore->delete_bucket(s->user.user_id, s->bucket_str);

    if (ret == 0) {
      rgw_remove_bucket(s->user.user_id, s->bucket_str, true);
    }
  }

  send_response();
}

struct put_obj_aio_info {
  void *data;
  void *handle;
};

static struct put_obj_aio_info pop_pending(std::list<struct put_obj_aio_info>& pending)
{
  struct put_obj_aio_info info;
  info = pending.front();
  pending.pop_front();
  return info;
}

static int wait_pending_front(std::list<struct put_obj_aio_info>& pending)
{
  struct put_obj_aio_info info = pop_pending(pending);
  int ret = rgwstore->aio_wait(info.handle);
  free(info.data);
  return ret;
}

static bool pending_has_completed(std::list<struct put_obj_aio_info>& pending)
{
  if (pending.size() == 0)
    return false;

  struct put_obj_aio_info& info = pending.front();
  return rgwstore->aio_completed(info.handle);
}

static int drain_pending(std::list<struct put_obj_aio_info>& pending)
{
  int ret = 0;
  while (!pending.empty()) {
    int r = wait_pending_front(pending);
    if (r < 0)
      ret = r;
  }
  return ret;
}

int RGWPutObj::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_WRITE))
    return -EACCES;

  return 0;
}

void RGWPutObj::execute()
{
  bool multipart;
  string multipart_meta_obj;
  string part_num;
  list<struct put_obj_aio_info> pending;
  size_t max_chunks = RGW_MAX_PENDING_CHUNKS;

  ret = -EINVAL;
  if (!s->object) {
    goto done;
  } else {
    ret = get_params();
    if (ret < 0)
      goto done;

    RGWAccessControlPolicy policy;

    ret = policy.create_canned(s->user.user_id, s->user.display_name, s->canned_acl);
    if (!ret) {
       ret = -EINVAL;
       goto done;
    }

    char supplied_md5_bin[CEPH_CRYPTO_MD5_DIGESTSIZE + 1];
    char supplied_md5[CEPH_CRYPTO_MD5_DIGESTSIZE * 2 + 1];
    char calc_md5[CEPH_CRYPTO_MD5_DIGESTSIZE * 2 + 1];
    unsigned char m[CEPH_CRYPTO_MD5_DIGESTSIZE];

    if (supplied_md5_b64) {
      RGW_LOG(15) << "supplied_md5_b64=" << supplied_md5_b64 << dendl;
      ret = ceph_unarmor(supplied_md5_bin, &supplied_md5_bin[CEPH_CRYPTO_MD5_DIGESTSIZE + 1],
			     supplied_md5_b64, supplied_md5_b64 + strlen(supplied_md5_b64));
      RGW_LOG(15) << "ceph_armor ret=" << ret << dendl;
      if (ret != CEPH_CRYPTO_MD5_DIGESTSIZE) {
        ret = -ERR_INVALID_DIGEST;
        goto done;
      }

      buf_to_hex((const unsigned char *)supplied_md5_bin, CEPH_CRYPTO_MD5_DIGESTSIZE, supplied_md5);
      RGW_LOG(15) << "supplied_md5=" << supplied_md5 << dendl;
    }

    MD5 hash;
    string oid;
    multipart = s->args.exists("uploadId");
    rgw_obj obj;
    if (!multipart) {
      oid = s->object_str;
      obj.set_ns(tmp_ns);

      char buf[33];
      gen_rand_alphanumeric(buf, sizeof(buf) - 1);
      oid.append("_");
      oid.append(buf);
    } else {
      oid = s->object_str;
      string upload_id = s->args.get("uploadId");
      RGWMPObj mp(oid, upload_id);
      multipart_meta_obj = mp.get_meta();

      part_num = s->args.get("partNumber");
      if (part_num.empty()) {
        ret = -EINVAL;
        goto done;
      }
      oid = mp.get_part(part_num);

      obj.set_ns(mp_ns);
    }
    obj.init(s->bucket_str, oid, s->object_str);
    do {
      get_data();
      if (len > 0) {
        struct put_obj_aio_info info;
        size_t orig_size;
	// For the first call to put_obj_data, pass -1 as the offset to
	// do a write_full.
        void *handle;
        ret = rgwstore->aio_put_obj_data(s->user.user_id, obj,
				     data,
				     ((ofs == 0) ? -1 : ofs), len, &handle);
        if (ret < 0)
          goto done;

        hash.Update((unsigned char *)data, len);
        info.handle = handle;
        info.data = data;
        pending.push_back(info);
        orig_size = pending.size();
        while (pending_has_completed(pending)) {
          ret = wait_pending_front(pending);
          if (ret < 0)
            goto done;

        }

        /* resize window in case messages are draining too fast */
        if (orig_size - pending.size() >= max_chunks)
          max_chunks++;

        if (pending.size() > max_chunks) {
          ret = wait_pending_front(pending);
          if (ret < 0)
            goto done;
        }
        ofs += len;
      }
    } while ( len > 0);
    drain_pending(pending);

    s->obj_size = ofs;

    hash.Final(m);

    buf_to_hex(m, CEPH_CRYPTO_MD5_DIGESTSIZE, calc_md5);

    if (supplied_md5_b64 && strcmp(calc_md5, supplied_md5)) {
       ret = -ERR_BAD_DIGEST;
       goto done;
    }
    bufferlist aclbl;
    policy.encode(aclbl);

    etag = calc_md5;
    map<string, bufferlist> attrs;
    bufferlist bl;
    bl.append(etag.c_str(), etag.size() + 1);
    attrs[RGW_ATTR_ETAG] = bl;
    attrs[RGW_ATTR_ACL] = aclbl;

    if (s->content_type) {
      bl.clear();
      bl.append(s->content_type, strlen(s->content_type) + 1);
      attrs[RGW_ATTR_CONTENT_TYPE] = bl;
    }

    get_request_metadata(s, attrs);

    if (!multipart) {
      rgw_obj dst_obj(s->bucket_str, s->object_str);
      ret = rgwstore->clone_obj(dst_obj, 0, obj, 0, s->obj_size, attrs);
      if (ret < 0)
        goto done;
      ret = rgwstore->delete_obj(s->user.user_id, obj);
      if (ret < 0)
        goto done;
    } else {
      ret = rgwstore->put_obj_meta(s->user.user_id, obj, NULL, attrs, false);
      if (ret < 0)
        goto done;

      bl.clear();
      map<string, bufferlist> meta_attrs;
      RGWUploadPartInfo info;
      string p = "part.";
      p.append(part_num);
      info.num = atoi(part_num.c_str());
      info.etag = etag;
      info.size = s->obj_size;
      info.modified = ceph_clock_now(g_ceph_context);
      ::encode(info, bl);
      meta_attrs[p] = bl;

      rgw_obj meta_obj(s->bucket_str, multipart_meta_obj, s->object_str, mp_ns);
      
      ret  = rgwstore->put_obj_meta(s->user.user_id, meta_obj, NULL, meta_attrs, false);
    }
  }
done:
  drain_pending(pending);
  send_response();
}

int RGWDeleteObj::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_WRITE))
    return -EACCES;

  return 0;
}

void RGWDeleteObj::execute()
{
  ret = -EINVAL;
  rgw_obj obj(s->bucket_str, s->object_str);
  if (s->object) {
    ret = rgwstore->delete_obj(s->user.user_id, obj);
  }

  send_response();
}

static bool parse_copy_source(const char *src, string& bucket, string& object)
{
  string url_src(src);
  string dec_src;

  url_decode(url_src, dec_src);
  src = dec_src.c_str();

  RGW_LOG(15) << "decoded src=" << src << dendl;

  if (*src == '/') ++src;

  string str(src);

  int pos = str.find("/");
  if (pos <= 0)
    return false;

  bucket = str.substr(0, pos);
  object = str.substr(pos + 1);

  if (object.size() == 0)
    return false;

  return true;
}

int RGWCopyObj::verify_permission()
{
  string empty_str;
  RGWAccessControlPolicy src_policy;
  RGWAccessControlPolicy dest_policy;

  if (!::verify_permission(s, RGW_PERM_WRITE))
    return -EACCES;

  ret = dest_policy.create_canned(s->user.user_id, s->user.display_name, s->canned_acl);
  if (!ret)
     return -EINVAL;

  ret = parse_copy_source(s->copy_source, src_bucket, src_object);
  if (!ret)
     return -EINVAL;

  /* just checking the bucket's permission */
  ret = read_acls(s, &src_policy, src_bucket, empty_str);
  if (ret < 0)
    return ret;

  if (!::verify_permission(&src_policy, s->user.user_id, s->perm_mask, RGW_PERM_READ))
    return -EACCES;

  dest_policy.encode(aclbl);

  return 0;
}


int RGWCopyObj::init_common()
{
  time_t mod_time;
  time_t unmod_time;
  time_t *mod_ptr = NULL;
  time_t *unmod_ptr = NULL;

  if (if_mod) {
    if (parse_time(if_mod, &mod_time) < 0) {
      ret = -EINVAL;
      return ret;
    }
    mod_ptr = &mod_time;
  }

  if (if_unmod) {
    if (parse_time(if_unmod, &unmod_time) < 0) {
      ret = -EINVAL;
      return ret;
    }
    unmod_ptr = &unmod_time;
  }

  attrs[RGW_ATTR_ACL] = aclbl;
  get_request_metadata(s, attrs);

  return 0;
}

void RGWCopyObj::execute()
{
  rgw_obj src_obj, dst_obj;

  ret = get_params();
  if (ret < 0)
    goto done;

  if (init_common() < 0)
    goto done;

  src_obj.init(src_bucket, src_object);
  dst_obj.init(s->bucket_str, s->object_str);
  ret = rgwstore->copy_obj(s->user.user_id,
                        dst_obj,
                        src_obj,
                        &mtime,
                        mod_ptr,
                        unmod_ptr,
                        if_match,
                        if_nomatch,
                        attrs, &s->err);

done:
  send_response();
}

int RGWGetACLs::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_READ_ACP))
    return -EACCES;

  return 0;
}

void RGWGetACLs::execute()
{
  ret = read_acls(s);

  if (ret < 0) {
    send_response();
    return;
  }

  stringstream ss;
  s->acl->to_xml(ss);
  acls = ss.str(); 

  send_response();
}

static int rebuild_policy(ACLOwner *owner, RGWAccessControlPolicy& src, RGWAccessControlPolicy& dest)
{
  if (!owner)
    return -EINVAL;

  ACLOwner *requested_owner = (ACLOwner *)src.find_first("Owner");
  if (requested_owner && requested_owner->get_id().compare(owner->get_id()) != 0) {
    return -EPERM;
  }

  RGWUserInfo owner_info;
  if (rgw_get_user_info_by_uid(owner->get_id(), owner_info) < 0) {
    RGW_LOG(10) << "owner info does not exist" << dendl;
    return -EINVAL;
  }
  ACLOwner& dest_owner = dest.get_owner();
  dest_owner.set_id(owner->get_id());
  dest_owner.set_name(owner_info.display_name);

  RGW_LOG(0) << "owner id=" << owner->get_id() << dendl;
  RGW_LOG(0) << "dest owner id=" << dest.get_owner().get_id() << dendl;

  RGWAccessControlList& src_acl = src.get_acl();
  RGWAccessControlList& acl = dest.get_acl();

  XMLObjIter iter = src_acl.find("Grant");
  ACLGrant *src_grant = (ACLGrant *)iter.get_next();
  while (src_grant) {
    ACLGranteeType& type = src_grant->get_type();
    ACLGrant new_grant;
    bool grant_ok = false;
    string id;
    RGWUserInfo grant_user;
    switch (type.get_type()) {
    case ACL_TYPE_EMAIL_USER:
      {
        string email = src_grant->get_id();
        RGW_LOG(10) << "grant user email=" << email << dendl;
        if (rgw_get_user_info_by_email(email, grant_user) < 0) {
          RGW_LOG(10) << "grant user email not found or other error" << dendl;
          return -ERR_UNRESOLVABLE_EMAIL;
        }
        id = grant_user.user_id;
      }
    case ACL_TYPE_CANON_USER:
      {
        if (type.get_type() == ACL_TYPE_CANON_USER)
          id = src_grant->get_id();
    
        if (grant_user.user_id.empty() && rgw_get_user_info_by_uid(id, grant_user) < 0) {
          RGW_LOG(10) << "grant user does not exist:" << id << dendl;
        } else {
          ACLPermission& perm = src_grant->get_permission();
          new_grant.set_canon(id, grant_user.display_name, perm.get_permissions());
          grant_ok = true;
          RGW_LOG(10) << "new grant: " << new_grant.get_id() << ":" << grant_user.display_name << dendl;
        }
      }
      break;
    case ACL_TYPE_GROUP:
      {
        string group = src_grant->get_id();
        if (group.compare(RGW_URI_ALL_USERS) == 0 ||
            group.compare(RGW_URI_AUTH_USERS) == 0) {
          new_grant = *src_grant;
          grant_ok = true;
          RGW_LOG(10) << "new grant: " << new_grant.get_id() << dendl;
        }
      }
    default:
      break;
    }
    if (grant_ok) {
      acl.add_grant(&new_grant);
    }
    src_grant = (ACLGrant *)iter.get_next();
  }

  return 0; 
}

int RGWPutACLs::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_WRITE_ACP))
    return -EACCES;

  return 0;
}

void RGWPutACLs::execute()
{
  bufferlist bl;

  RGWAccessControlPolicy *policy = NULL;
  RGWACLXMLParser parser;
  RGWAccessControlPolicy new_policy;
  stringstream ss;
  char *orig_data = data;
  char *new_data = NULL;
  ACLOwner owner;
  rgw_obj obj;

  ret = 0;

  if (!parser.init()) {
    ret = -EINVAL;
    goto done;
  }

  if (!s->acl) {
     s->acl = new RGWAccessControlPolicy;
     if (!s->acl) {
       ret = -ENOMEM;
       goto done;
     }
     owner.set_id(s->user.user_id);
     owner.set_name(s->user.display_name);
  } else {
     owner = s->acl->get_owner();
  }

  if (get_params() < 0)
    goto done;

  RGW_LOG(15) << "read len=" << len << " data=" << (data ? data : "") << dendl;

  if (!s->canned_acl.empty() && len) {
    ret = -EINVAL;
    goto done;
  }
  if (!s->canned_acl.empty()) {
    RGWAccessControlPolicy canned_policy;
    bool r = canned_policy.create_canned(owner.get_id(), owner.get_display_name(), s->canned_acl);
    if (!r) {
      ret = -EINVAL;
      goto done;
    }
    canned_policy.to_xml(ss);
    new_data = strdup(ss.str().c_str());
    data = new_data;
    len = ss.str().size();
  }


  if (!parser.parse(data, len, 1)) {
    ret = -EACCES;
    goto done;
  }
  policy = (RGWAccessControlPolicy *)parser.find_first("AccessControlPolicy");
  if (!policy) {
    ret = -EINVAL;
    goto done;
  }

  if (g_conf->rgw_log >= 15) {
    RGW_LOG(15) << "Old AccessControlPolicy" << dendl;
    policy->to_xml(cout);
    RGW_LOG(15) << dendl;
  }

  ret = rebuild_policy(&owner, *policy, new_policy);
  if (ret < 0)
    goto done;

  if (g_conf->rgw_log >= 15) {
    RGW_LOG(15) << "New AccessControlPolicy" << dendl;
    new_policy.to_xml(cout);
    RGW_LOG(15) << dendl;
  }

  new_policy.encode(bl);
  obj.init(s->bucket_str, s->object_str);
  ret = rgwstore->set_attr(obj, RGW_ATTR_ACL, bl);

done:
  free(orig_data);
  free(new_data);

  send_response();
}

int RGWInitMultipart::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_WRITE))
    return -EACCES;

  return 0;
}

void RGWInitMultipart::execute()
{
  bufferlist bl;
  bufferlist aclbl;
  RGWAccessControlPolicy policy;
  map<string, bufferlist> attrs;
  rgw_obj obj;

  if (get_params() < 0)
    goto done;
  ret = -EINVAL;
  if (!s->object)
    goto done;

  ret = policy.create_canned(s->user.user_id, s->user.display_name, s->canned_acl);
  if (!ret) {
     ret = -EINVAL;
     goto done;
  }

  policy.encode(aclbl);

  attrs[RGW_ATTR_ACL] = aclbl;

  if (s->content_type) {
    bl.append(s->content_type, strlen(s->content_type) + 1);
    attrs[RGW_ATTR_CONTENT_TYPE] = bl;
  }

  get_request_metadata(s, attrs);

  do {
    char buf[33];
    gen_rand_alphanumeric(buf, sizeof(buf) - 1);
    upload_id = buf;

    string tmp_obj_name;
    RGWMPObj mp(s->object_str, upload_id);
    tmp_obj_name = mp.get_meta();

    obj.init(s->bucket_str, tmp_obj_name, s->object_str, mp_ns);
    ret = rgwstore->put_obj_meta(s->user.user_id, obj, NULL, attrs, true);
  } while (ret == -EEXIST);
done:
  send_response();
}

static int get_multiparts_info(struct req_state *s, string& meta_oid, map<uint32_t, RGWUploadPartInfo>& parts,
                               RGWAccessControlPolicy& policy, map<string, bufferlist>& new_attrs)
{
  void *handle;
  map<string, bufferlist> attrs;
  map<string, bufferlist>::iterator iter;

  rgw_obj obj(s->bucket_str, meta_oid, s->object_str, mp_ns);

  int ret = rgwstore->prepare_get_obj(obj, 0, NULL, &attrs, NULL,
                                      NULL, NULL, NULL, NULL, NULL, &handle, &s->err);
  rgwstore->finish_get_obj(&handle);

  if (ret < 0)
    return ret;

  for (iter = attrs.begin(); iter != attrs.end(); ++iter) {
    string name = iter->first;
    if (name.compare(RGW_ATTR_ACL) == 0) {
      bufferlist& bl = iter->second;
      bufferlist::iterator bli = bl.begin();
      ::decode(policy, bli);
      new_attrs[RGW_ATTR_ACL] = bl;
      continue;
    }
    if (name.compare(0, 5, "part.") != 0) {
      new_attrs[iter->first] = iter->second;
      continue;
    }

    bufferlist& bl = iter->second;
    bufferlist::iterator bli = bl.begin();
    RGWUploadPartInfo info;
    ::decode(info, bli);
    parts[info.num] = info;
  }
  return 0;
}

int RGWCompleteMultipart::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_WRITE))
    return -EACCES;

  return 0;
}

void RGWCompleteMultipart::execute()
{
  RGWMultiCompleteUpload *parts;
  map<int, string>::iterator iter;
  RGWMultiXMLParser parser;
  string meta_oid;
  map<uint32_t, RGWUploadPartInfo> obj_parts;
  map<uint32_t, RGWUploadPartInfo>::iterator obj_iter;
  RGWAccessControlPolicy policy;
  map<string, bufferlist> attrs;
  off_t ofs = 0;
  MD5 hash;
  char final_etag[CEPH_CRYPTO_MD5_DIGESTSIZE];
  char final_etag_str[CEPH_CRYPTO_MD5_DIGESTSIZE * 2 + 16];
  bufferlist etag_bl;
  rgw_obj meta_obj;
  rgw_obj target_obj;
  RGWMPObj mp;
  vector<RGWCloneRangeInfo> ranges;


  ret = get_params();
  if (ret < 0)
    goto done;

  if (!data) {
    ret = -EINVAL;
    goto done;
  }

  if (!parser.init()) {
    ret = -EINVAL;
    goto done;
  }

  if (!parser.parse(data, len, 1)) {
    ret = -EINVAL;
    goto done;
  }

  parts = (RGWMultiCompleteUpload *)parser.find_first("CompleteMultipartUpload");
  if (!parts) {
    ret = -EINVAL;
    goto done;
  }

  mp.init(s->object_str, upload_id);
  meta_oid = mp.get_meta();

  ret = get_multiparts_info(s, meta_oid, obj_parts, policy, attrs);
  if (ret == -ENOENT)
    ret = -ERR_NO_SUCH_UPLOAD;
  if (parts->parts.size() != obj_parts.size())
    ret = -ERR_INVALID_PART;
  if (ret < 0)
    goto done;

  for (iter = parts->parts.begin(), obj_iter = obj_parts.begin();
       iter != parts->parts.end() && obj_iter != obj_parts.end();
       ++iter, ++obj_iter) {
    char etag[CEPH_CRYPTO_MD5_DIGESTSIZE];
    if (iter->first != (int)obj_iter->first) {
      RGW_LOG(0) << "parts num mismatch: next requested: " << iter->first << " next uploaded: " << obj_iter->first << dendl;
      ret = -ERR_INVALID_PART;
      goto done;
    }
    if (iter->second.compare(obj_iter->second.etag) != 0) {
      RGW_LOG(0) << "part: " << iter->first << " etag: " << iter->second << dendl;
      ret = -ERR_INVALID_PART;
      goto done;
    }

    hex_to_buf(obj_iter->second.etag.c_str(), etag, CEPH_CRYPTO_MD5_DIGESTSIZE);
    hash.Update((const byte *)etag, sizeof(etag));
  }
  hash.Final((byte *)final_etag);

  buf_to_hex((unsigned char *)final_etag, sizeof(final_etag), final_etag_str);
  snprintf(&final_etag_str[CEPH_CRYPTO_MD5_DIGESTSIZE * 2],  sizeof(final_etag_str) - CEPH_CRYPTO_MD5_DIGESTSIZE * 2,
           "-%lld", (long long)parts->parts.size());
  RGW_LOG(0) << "calculated etag: " << final_etag_str << dendl;

  etag_bl.append(final_etag_str, strlen(final_etag_str) + 1);

  attrs[RGW_ATTR_ETAG] = etag_bl;

  target_obj.init(s->bucket_str, s->object_str);
  ret = rgwstore->put_obj_meta(s->user.user_id, target_obj, NULL, attrs, false);
  if (ret < 0)
    goto done;
  
  for (obj_iter = obj_parts.begin(); obj_iter != obj_parts.end(); ++obj_iter) {
    string oid = mp.get_part(obj_iter->second.num);
    rgw_obj src_obj(s->bucket_str, oid, s->object_str, mp_ns);

    RGWCloneRangeInfo range;
    range.src = src_obj;
    range.src_ofs = 0;
    range.dst_ofs = ofs;
    range.len = obj_iter->second.size;
    ranges.push_back(range);

    ofs += obj_iter->second.size;
  }
  ret = rgwstore->clone_objs(target_obj, ranges, attrs, true);
  if (ret < 0)
    goto done;

  // now erase all parts
  for (obj_iter = obj_parts.begin(); obj_iter != obj_parts.end(); ++obj_iter) {
    string oid = mp.get_part(obj_iter->second.num);
    rgw_obj obj(s->bucket_str, oid, s->object_str, mp_ns);
    rgwstore->delete_obj(s->user.user_id, obj);
  }
  // and also remove the metadata obj
  meta_obj.init(s->bucket_str, meta_oid, s->object_str, mp_ns);
  rgwstore->delete_obj(s->user.user_id, meta_obj);

done:
  send_response();
}

int RGWAbortMultipart::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_WRITE))
    return -EACCES;

  return 0;
}

void RGWAbortMultipart::execute()
{
  ret = -EINVAL;
  string upload_id = s->args.get("uploadId");
  string meta_oid;
  string prefix;
  map<uint32_t, RGWUploadPartInfo> obj_parts;
  map<uint32_t, RGWUploadPartInfo>::iterator obj_iter;
  RGWAccessControlPolicy policy;
  map<string, bufferlist> attrs;
  rgw_obj meta_obj;
  RGWMPObj mp;

  if (upload_id.empty() || s->object_str.empty())
    goto done;

  mp.init(s->object_str, upload_id); 
  meta_oid = mp.get_meta();

  ret = get_multiparts_info(s, meta_oid, obj_parts, policy, attrs);
  if (ret < 0)
    goto done;

  for (obj_iter = obj_parts.begin(); obj_iter != obj_parts.end(); ++obj_iter) {
    string oid = mp.get_part(obj_iter->second.num);
    rgw_obj obj(s->bucket_str, oid, s->object_str, mp_ns);
    ret = rgwstore->delete_obj(s->user.user_id, obj);
    if (ret < 0 && ret != -ENOENT)
      goto done;
  }
  // and also remove the metadata obj
  meta_obj.init(s->bucket_str, meta_oid, s->object_str, mp_ns);
  ret = rgwstore->delete_obj(s->user.user_id, meta_obj);
  if (ret == -ENOENT) {
    ret = -ERR_NO_SUCH_BUCKET;
  }
done:

  send_response();
}

int RGWListMultipart::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_READ))
    return -EACCES;

  return 0;
}

void RGWListMultipart::execute()
{
  map<string, bufferlist> xattrs;
  string meta_oid;
  RGWMPObj mp;

  ret = get_params();
  if (ret < 0)
    goto done;

  mp.init(s->object_str, upload_id);
  meta_oid = mp.get_meta();

  ret = get_multiparts_info(s, meta_oid, parts, policy, xattrs);

done:
  send_response();
}

int RGWListBucketMultiparts::verify_permission()
{
  if (!::verify_permission(s, RGW_PERM_READ))
    return -EACCES;

  return 0;
}

void RGWListBucketMultiparts::execute()
{
  vector<RGWObjEnt> objs;
  string marker_meta;

  ret = get_params();
  if (ret < 0)
    goto done;

  if (s->prot_flags & RGW_REST_OPENSTACK) {
    string path_args = s->args.get("path");
    if (!path_args.empty()) {
      if (!delimiter.empty() || !prefix.empty()) {
        ret = -EINVAL;
        goto done;
      }
      url_decode(path_args, prefix);
      delimiter="/";
    }
  }
  marker_meta = marker.get_meta();
  ret = rgwstore->list_objects(s->user.user_id, s->bucket_str, max_uploads, prefix, delimiter, marker_meta, objs, common_prefixes,
                               !!(s->prot_flags & RGW_REST_OPENSTACK), mp_ns, &is_truncated, &mp_filter);
  if (objs.size()) {
    vector<RGWObjEnt>::iterator iter;
    RGWMultipartUploadEntry entry;
    for (iter = objs.begin(); iter != objs.end(); ++iter) {
      string name = iter->name;
      if (!entry.mp.from_meta(name))
        continue;
      entry.obj = *iter;
      uploads.push_back(entry);
    }
    next_marker = entry;
  }
done:
  send_response();
}

void RGWHandler::init_state(struct req_state *s, struct fcgx_state *fcgx)
{
  /* Retrieve the loglevel from the CGI envirioment (if set) */
  const char *cgi_env_level = FCGX_GetParam("RGW_LOG_LEVEL", fcgx->envp);
  if (cgi_env_level != NULL) {
    int level = atoi(cgi_env_level);
    if (level >= 0) {
      g_conf->rgw_log = level;
    }
  }

  const char *cgi_should_log = FCGX_GetParam("RGW_SHOULD_LOG", fcgx->envp);
  s->should_log = rgw_str_to_bool(cgi_should_log, RGW_SHOULD_LOG_DEFAULT);

  if (g_conf->rgw_log >= 20) {
    char *p;
    for (int i=0; (p = fcgx->envp[i]); ++i) {
      RGW_LOG(20) << p << dendl;
    }
  }
  s->fcgx = fcgx;
  s->content_started = false;
  s->err.clear();
  s->format = 0;
  if (s->acl) {
     delete s->acl;
     s->acl = new RGWAccessControlPolicy;
  }
  s->canned_acl.clear();
  s->expect_cont = false;

  free(s->os_user);
  free(s->os_groups);
  s->os_auth_token = NULL;
  s->os_user = NULL;
  s->os_groups = NULL;
  s->time = ceph_clock_now(g_ceph_context);
  s->user.clear();
  s->perm_mask = 0;
}

int RGWHandler::do_read_permissions(bool only_bucket)
{
  int ret = read_acls(s, only_bucket);

  if (ret < 0) {
    RGW_LOG(10) << "read_permissions on " << s->bucket_str << ":" <<s->object_str << " only_bucket=" << only_bucket << " ret=" << ret << dendl;
    if (ret == -ENODATA)
      ret = -EACCES;
  }

  if (!s->bucket_str.empty())
    s->pool_id = rgwstore->get_bucket_id(s->bucket_str);

  return ret;
}

